# frozen_string_literal: true

module ErgomentumRspec
  module Matchers
    module Performance
      def allocates
        AllocateObjects.new
      end

      class IncompleteExpectationError < StandardError
        def initialize
          super("Comparision operator required. E. g. allocates.exactly(1)")
        end
      end

      class AllocateObjects
        def supports_block_expectations?
          true
        end

        def matches?(*actual, &block)
          @trace = if block
                     AllocationStats.trace(&block)
                   elsif actual.first.is_a?(Proc)
                     AllocationStats.trace(&actual.first)
                   else
                     actual.first
                   end
          @compare || raise(IncompleteExpectationError)
          @compare.call
        end

        def exactly(expected)
          @expected = expected
          @compare = proc { @trace.new_allocations.size == @expected }
          @message = ""
          self
        end

        def less_than(expected)
          @expected = expected
          @compare = proc { @trace.new_allocations.size < @expected }
          @message = "less than "
          self
        end

        def more_than(expected)
          @expected = expected
          @compare = proc { @trace.new_allocations.size > @expected }
          @message = "more than "
          self
        end

        def max(expected)
          @expected = expected
          @compare = proc { @trace.new_allocations.size <= @expected }
          @message = "max "
          self
        end

        def min(expected)
          @expected = expected
          @compare = proc { @trace.new_allocations.size >= @expected }
          @message = "min "
          self
        end

        def object
          self
        end

        def objects
          self
        end

        def failure_message
          "expected #{@message}#{@expected} objects to be allocated; "\
            "got #{@trace.new_allocations.size}:\n\n"\
            "#{@trace.allocations(alias_paths: true).group_by(:sourcefile, :sourceline, :class).to_text}"
        end

        def failure_message_when_negated
          "expected not #{@message}#{@expected} objects to be allocated; "\
            "got #{@trace.new_allocations.size}:\n\n"\
            "#{@trace.allocations(alias_paths: true).group_by(:sourcefile, :sourceline, :class).to_text}"
        end

        def description
          "allocates number of objects"
        end
      end
    end
  end
end
